import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;

import javax.imageio.ImageIO;
import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.JLabel;

import org.opencv.core.Core;
import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.MatOfByte;
import org.opencv.core.MatOfPoint;
import org.opencv.core.Point;
import org.opencv.core.Rect;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.highgui.Highgui;
import org.opencv.highgui.VideoCapture;
import org.opencv.imgproc.Imgproc;
// Tracking a Red Circular Item for Augustin Ubbo Bot
// By Justin CANO, Ecole Centrale de Marseille, 13-06-2016
// Contact cano.justin@gmail.com
// Tracking code adapted from https://ratiler.wordpress.com/2014/09/21/detection-et-suivi-dun-objet-colore/


public class Tracking {

// First, we need a Tracking Class that we will include in the Ubbo's application source code 
	
	protected String command = ""; // The command
	protected String previousCommand = "NULL"; // Previous command
	protected double x; // Cartesian coordinates of the red spoted item
	protected double y;
	protected double area;
	protected boolean graphic; // If it's false, we wont't have any graphic frame (and save memory so far)
	protected boolean trackOk; // Enable/Disable tracking to go out the routine loop
	protected int success; // Number of tracking consecutive command success (we need N similar tracking command to get the command actually applied)
	
	static {
		System.loadLibrary(Core.NATIVE_LIBRARY_NAME);
	}
	static Mat frame = null;

	public Tracking() {
		// Default constructor
		command = ""; // No command for the begining
		previousCommand = "";
		x = 300; // Point corresponding approx. to the center of the image...   
		y = 160; // and...
		area = 20000; // ...the item is enough close to the webcam
		// so, the Tracking will return a NULL command (red spot image perfectly centered)
		graphic = false;
		trackOk = false; // Caution ! You must enable tracking (TrackOk) and graphics (graphic) in view to set your tracking algoritm
		success = 0; // Number of success initialized to 0 
	}

	public void track() {
	    // Main method
		if( !trackOk) {
			System.out.println("Tracking Disabled"); // No tracking allowed
		}
		else {
			if(this.graphic) {
				this.trackingWithGraphic(); // Mode with graphic JFrame
			}
			else {
				this.trackingWithoutGraphic(); // Without 
			}
		
		}
	}

	public void trackingWithGraphic() {
		// Tracking Routine with Graphic (J Frame with video return) (Adapted from : https://ratiler.wordpress.com/2014/09/21/detection-et-suivi-dun-objet-colore/ )

		JFrame jframe = new JFrame("Augustin Tracking");
		jframe.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		JLabel vidpanel = new JLabel();
		jframe.setContentPane(vidpanel);
		jframe.setSize(640, 480);
		jframe.setVisible(true);

		VideoCapture capture = new VideoCapture(0);

		Mat frame = new Mat();
		Mat hsv_image = new Mat();
		Mat thresholded = new Mat();
		Mat thresholded2 = new Mat();
		Mat circles = new Mat();

		Mat array255 = new Mat(480, 640, CvType.CV_8UC1);
		array255.setTo(new Scalar(255));

		Mat distance = new Mat(480, 640, CvType.CV_8UC1);
		List<Mat> lhsv = new ArrayList<Mat>(3);
		Scalar hsv_min = new Scalar(0, 50, 50, 0);
		Scalar hsv_max = new Scalar(6, 255, 255, 0);
		Scalar hsv_min2 = new Scalar(175, 50, 50, 0);
		Scalar hsv_max2 = new Scalar(179, 255, 255, 0);

		Size sz = new Size(640, 480);

		capture.read(frame);


		if (capture.isOpened()) {
			//System.out.println("Frame Ouverte");
			while (trackOk) {
				capture.read(frame);
				if (!frame.empty()) {
					//System.out.println("Frame Non Vide");

					Imgproc.resize(frame, frame, sz);

					Imgproc.cvtColor(frame, hsv_image, Imgproc.COLOR_BGR2HSV);
					Core.inRange(hsv_image, hsv_min, hsv_max, thresholded);
					Core.inRange(hsv_image, hsv_min2, hsv_max2, thresholded2);
					Core.bitwise_or(thresholded, thresholded2, thresholded);

					Core.split(hsv_image, lhsv);

					Mat S = lhsv.get(1);
					Mat V = lhsv.get(2);

					Core.subtract(array255, S, S);
					Core.subtract(array255, V, V);

					S.convertTo(S, CvType.CV_32F);
					V.convertTo(V, CvType.CV_32F);

					Core.magnitude(S, V, distance);

					Core.inRange(distance, new Scalar(0.0), new Scalar(200.0),thresholded2);

					Core.bitwise_and(thresholded, thresholded2, thresholded);
					Imgproc.GaussianBlur(thresholded, thresholded, new Size(9,9), 0, 0);
					Imgproc.HoughCircles(thresholded, circles,
							Imgproc.CV_HOUGH_GRADIENT, 2,
							thresholded.height() / 4, 500, 50, 0, 0);
					int rows = circles.rows();
					int elemSize = (int) circles.elemSize();
					float[] data2 = new float[rows * elemSize / 4];
					if (data2.length > 0) {
						circles.get(0, 0, data2);
						for (int i = 0; i < data2.length; i = i + 3) {
							Point center = new Point(data2[i], data2[i + 1]);
							Core.ellipse(frame, center, new Size(
									(double) data2[i + 2],
									(double) data2[i + 2]), 0, 0, 360,
									new Scalar(255, 0, 255), 4, 8, 0);
						}
						Rect r = detect_red_ball(thresholded);
						if (r != null) {
							Core.rectangle(frame, r.tl(), r.br(), new Scalar(0,
									255, 0), 2);
						}
						Core.putText(frame,"( x = "+0.5*(r.tl().x+r.br().x)+", y = "+0.5*(r.tl().y+r.br().y)+" )",new Point(0.5*(r.tl().x+r.br().x),0.5*(r.tl().y+r.br().y)),1,1,new Scalar(255,255,255)) ;
						x = r.br().x;
						y = r.br().y;
						double area = (r.width)*(r.height);
						System.out.println("x =" + x + "  y =" + y + "  area =" + area);
						corrections(x, y, area);
						System.out.println(this.getCommand());
					
						
						
						//System.out.println(this.command);
					}


					ImageIcon image = new ImageIcon(Mat2bufferedImage(frame));
					vidpanel.setIcon(image);
					vidpanel.repaint();


				}
				trackingFollowed(); // Check if the tracking is followed or not (external calling)
			}
		}
		
	}
	public void trackingWithoutGraphic() {
		// "Light" Tracking without JFrame (adapted from the https://ratiler.wordpress.com/2014/09/21/detection-et-suivi-dun-objet-colore/ code)

		VideoCapture capture = new VideoCapture(0);

		Mat frame = new Mat();
		Mat hsv_image = new Mat();
		Mat thresholded = new Mat();
		Mat thresholded2 = new Mat();
		Mat circles = new Mat();

		Mat array255 = new Mat(480, 640, CvType.CV_8UC1);
		array255.setTo(new Scalar(255));

		Mat distance = new Mat(480, 640, CvType.CV_8UC1);
		List<Mat> lhsv = new ArrayList<Mat>(3);
		Scalar hsv_min = new Scalar(0, 50, 50, 0);
		Scalar hsv_max = new Scalar(6, 255, 255, 0);
		Scalar hsv_min2 = new Scalar(175, 50, 50, 0);
		Scalar hsv_max2 = new Scalar(179, 255, 255, 0);

		Size sz = new Size(640, 480);

		capture.read(frame);


		if (capture.isOpened()) {
			//System.out.println("Frame Ouverte");
			while (trackOk) {
				capture.read(frame);
				if (!frame.empty()) {
					//System.out.println("Frame Non Vide");

					Imgproc.resize(frame, frame, sz);

					Imgproc.cvtColor(frame, hsv_image, Imgproc.COLOR_BGR2HSV);
					Core.inRange(hsv_image, hsv_min, hsv_max, thresholded);
					Core.inRange(hsv_image, hsv_min2, hsv_max2, thresholded2);
					Core.bitwise_or(thresholded, thresholded2, thresholded);

					Core.split(hsv_image, lhsv);

					Mat S = lhsv.get(1);
					Mat V = lhsv.get(2);

					Core.subtract(array255, S, S);
					Core.subtract(array255, V, V);

					S.convertTo(S, CvType.CV_32F);
					V.convertTo(V, CvType.CV_32F);

					Core.magnitude(S, V, distance);

					Core.inRange(distance, new Scalar(0.0), new Scalar(200.0),thresholded2);

					Core.bitwise_and(thresholded, thresholded2, thresholded);
					Imgproc.GaussianBlur(thresholded, thresholded, new Size(9,9), 0, 0);
					Imgproc.HoughCircles(thresholded, circles,
							Imgproc.CV_HOUGH_GRADIENT, 2,
							thresholded.height() / 4, 500, 50, 0, 0);
					int rows = circles.rows();
					int elemSize = (int) circles.elemSize();
					float[] data2 = new float[rows * elemSize / 4];
					if (data2.length > 0) {
						circles.get(0, 0, data2);
						for (int i = 0; i < data2.length; i = i + 3) {
							Point center = new Point(data2[i], data2[i + 1]);
							Core.ellipse(frame, center, new Size(
									(double) data2[i + 2],
									(double) data2[i + 2]), 0, 0, 360,
									new Scalar(255, 0, 255), 4, 8, 0);
						}
						Rect r = detect_red_ball(thresholded);
						if (r != null) {
							Core.rectangle(frame, r.tl(), r.br(), new Scalar(0,
									255, 0), 2);
						}
						Core.putText(frame,"( x = "+0.5*(r.tl().x+r.br().x)+", y = "+0.5*(r.tl().y+r.br().y)+" )",new Point(0.5*(r.tl().x+r.br().x),0.5*(r.tl().y+r.br().y)),1,1,new Scalar(255,255,255)) ;
						x = r.br().x;
						y = r.br().y;
						double area = (r.width)*(r.height);
						System.out.println("x =" + x + "  y =" + y + "  area =" + area);
						corrections(x, y, area);
						System.out.println(this.getCommand());
					
						
						
						//System.out.println(this.command);
					}

					//System.out
	


				}
				trackingFollowed(); // Check if the tracking is followed or not (external calling)
			}
		}
		
	}

// The two following method are used by the tracking algorithm (I have not modified either of those)
	public static BufferedImage Mat2bufferedImage(Mat image) {
		MatOfByte bytemat = new MatOfByte();
		Highgui.imencode(".jpg", image, bytemat);
		byte[] bytes = bytemat.toArray();
		InputStream in = new ByteArrayInputStream(bytes);
		BufferedImage img = null;
		try {
			img = ImageIO.read(in);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return img;
	}

	public static Rect detect_red_ball(Mat outmat) {

		Mat v = new Mat();
		List<MatOfPoint> contours = new ArrayList<MatOfPoint>();
		Imgproc.findContours(outmat, contours, v, Imgproc.RETR_LIST,
				Imgproc.CHAIN_APPROX_SIMPLE);

		double maxArea = -1;
		int maxAreaIdx = -1;
		Rect r = null;

		for (int idx = 0; idx < contours.size(); idx++) {
			Mat contour = contours.get(idx);
			double contourarea = Imgproc.contourArea(contour);
			if (contourarea > maxArea) {
				maxArea = contourarea;
				maxAreaIdx = idx;
				r = Imgproc.boundingRect(contours.get(maxAreaIdx));
			}

		}

		v.release();

		return r;

	}

// Those three following methods are returning booleans to mention if 
// correction have been used of not

	public boolean correction_rotation(double x) {
		// First, we will correct the angular position of the robot in the (x,y) plan
		// For this, we have two threesold in the x axis, x needs to be on this range
		double x_low = 250.0;
		double x_high = 380.0;
		boolean b = false;
		if(x_low > x) {
			this.command = "toRIGHT";
			b = true;
		}
		else if(x_high < x) {
			this.command = "toLEFT";
			b = true;
		}
		else {
			this.command = "NULL";
		}

		return b;
	}

	public boolean correction_translation(double area) {
		// If the area of the circular object is too little then the item is too far
		// -> The robots needs to approach the object
		double proximity_area = 15000.0;
		boolean b = false;
		if(proximity_area>area) {
			this.command = "toFORWARD";
			b = true;
		}
		else {
			this.command = "NULL";
		}
		return b;
	}

	public boolean correction_inclinaison(double y) {
		// idem than correction_translation() but here for the y axis
		double y_low = 250.0;
		double y_high = 350.0;
		boolean b = false;
		if(y_low > y) {
			this.command = "toDOWN";

			b = true;
		}
		else if(y_high < y) {
			this.command = "toTOP";
			b = true;
		}
		else {
			this.command = "NULL";
		}

		return b;
	}

	public void corrections (double x, double y, double d) {
		// We are going to correct only one parameter per iteration
		// 1 : x axis
		// 2 : y axis
		// 3 : z axis (area, approach of the object)
		
		boolean a = correction_rotation(x);
		if(!a) {
			boolean b = correction_inclinaison(y);
			if(!b){
				correction_translation(d);
			}
		}
	}

	public String getCommand() {
		// To get the command is compulsory to determine if the command is feasible 
		// (see next method)
		String s;
		if(this.checkCommand()) {
			s = this.command; 
		}
		else {
			s = "NULL";
		}
		return s;
	}

	public boolean checkCommand() {
		// If 5 attempted commands are the same, then your tracking suceed this test
		// This test is coded to avoid noise effect (tracking errors) as a low pass filter
		boolean b = false;
		if(this.previousCommand.equals(this.command)) {
			this.success++;
		}
		else {
			this.success = 0;
		}
		if(this.success>4){
			b = true;
		}
		else {
			b = false;
		}
		
		System.out.println("command precedente" + this.previousCommand);
		this.previousCommand = this.command;
		return b;
	}
	
	public void trackingFollowed() {
		if(this.command.equals("toTOP")){
			//this.trackOk = false;
			//System.out.println("Sortie du process");
		}
	}
	
	// Two methods for external command of tracking
	public void enableTrack() {
		this.trackOk = true;
	}
	public void disableTrack() {
		this.trackOk = false;
	}
	
	// Two methods to enable/disable graphics
	public void enableGraphic() {
		this.graphic = true;
	}
	public void disableGraphic () {
		this.graphic = false;
	}

}
